/******************************************************************************
 *
 * Name:     gdal_vector_cpp.h
 * Project:  GDAL Core
 * Purpose:  GDAL Core Vector C++ declarations.
 * Author:   Even Rouault, <even dot rouault at spatialys.com>
 *
 ******************************************************************************
 * Copyright (c) 2025 Even Rouault <even dot rouault at spatialys.com>
 *
 * SPDX-License-Identifier: MIT
 ****************************************************************************/

#ifndef GDAL_VECTOR_CPP_H_INCLUDED
#define GDAL_VECTOR_CPP_H_INCLUDED

/**
 * \file gdal_vector_cpp.h
 *
 * C++ GDAL vector API entry points.
 *
 * \since GDAL 3.12
 */

#include "gdal_dataset.h"
#include "gdal_drivermanager.h"
#include "ogrsf_frmts.h"
#include "ogr_feature.h"
#include "ogr_geometry.h"

#endif
